"""https://www.notion.so/castordoc/Workshop-Mocking-405ef02712e6446193720abf8d4c2f53"""

from unittest.mock import patch

from ..abstract import SqlalchemyClient
from .client import SnowflakeClient, _scalar


class MockedResult:
    def scalar(self):
        return "foo"


class MockedConnection:
    def execute(self, query):
        return MockedResult()


def test_scalar():
    mocked_connection = MockedConnection()
    result = _scalar(mocked_connection, "SELECT foo FROM BAR;")
    assert result == "foo"


def get_snowflake_connection():
    credentials = {
        "account": "value_account",
        "user": "value_user",
        "password": "value_password",
    }
    return SnowflakeClient(credentials)


@patch.object(SqlalchemyClient, "__init__")
def test_build_uri(_):
    credentials = {
        "account": "value_account",
        "user": "value_user",
        "password": "value_password",
    }
    client = get_snowflake_connection()
    result = client._build_uri(credentials)
    assert (
        result
        == "snowflake://value_user:value_password@value_account/?application=castor"
    )


@patch.object(SqlalchemyClient, "__init__")
@patch(
    "source.packages.extractor.castor_extractor.warehouse.snowflake.client._use",
)
def test_role(mocked_used, _):
    client = get_snowflake_connection()
    client._role = "ACCOUNTADMIN"
    client.role(MockedConnection())
    mocked_used.assert_called_once()



from typing import Dict, List, Tuple, Union

from tableauserverclient import ServerResponseError  # type: ignore
from typing_extensions import Literal

from .errors import TableauErrorCode

PageReturn = Union[
    Tuple[List[Dict], Literal[None]],
    Tuple[Literal[None], Union[TableauErrorCode, ServerResponseError]],
]



import datetime
from unittest.mock import patch

import pytest
from castor_extractor.visualization.looker import (  # type: ignore
    ApiClient,
    Credentials,
)
from dateutil.utils import today
from freezegun import freeze_time
from source.packages.extractor.castor_extractor.visualization.looker.api.client import (
    _mondays,
)


def _credentials():
    return Credentials(  # noqa: S106
        base_url="base_url",
        client_id="client_id",
        client_secret="secret",
    )


@patch("castor_extractor.visualization.looker.api.client.init40")
@patch("castor_extractor.visualization.looker.api.client.has_admin_permissions")
def test_api_client_has_admin_permissions(
    mock_has_admin_permission,
    mock_init40,
):
    mock_has_admin_permission.return_value = False
    with pytest.raises(PermissionError):
        ApiClient(_credentials())

    mock_has_admin_permission.return_value = True
    mock_init40.return_value = "sdk"
    client = ApiClient(_credentials())
    assert client._sdk == "sdk"


@freeze_time("2023-7-4")
def test__mondays():
    expected_30_days = {
        datetime.date(2023, 7, 3),
        datetime.date(2023, 6, 26),
        datetime.date(2023, 6, 19),
        datetime.date(2023, 6, 12),
        datetime.date(2023, 6, 5),
    }
    assert set(_mondays(history_depth_in_days=30)) == expected_30_days

    # only mondays
    assert all([day.weekday() == 0 for day in _mondays(30)])
    assert all([day.weekday() == 0 for day in _mondays(100)])

    # all days must remain in the elapsed history
    history_days = 1_000
    end = today().date()
    start = end - datetime.timedelta(days=history_days)
    output = _mondays(history_depth_in_days=history_days)

    assert all([day >= start for day in output])
    assert all([day < end for day in output])

    with pytest.raises(AssertionError):
        list(_mondays(history_depth_in_days=0))
        list(_mondays(history_depth_in_days=-5))



import logging
from enum import Enum
from typing import Optional

import requests

from .credentials import CredentialsKey, DbtCredentials, get_value

logger = logging.getLogger(__name__)

_CLOUD_URL = "https://cloud.getdbt.com/api/v2/accounts"
_DATA_KEY = "data"
_MANIFEST_PATH = "manifest.json"
_SUCCESSFUL_RUN_STATUS = 10


class ContentType(Enum):
    JSON = "application/json"
    TEXT = "text/html"


class DbtClient:
    """
    Connect to dbt-cloud Administrative API and downloads manifest.
    https://docs.getdbt.com/docs/dbt-cloud-apis/admin-cloud-api
    """

    def __init__(self, **kwargs):
        self._credentials = DbtCredentials(
            token=get_value(CredentialsKey.TOKEN, kwargs),
            job_id=get_value(CredentialsKey.JOB_ID, kwargs),
        )
        self._session = requests.Session()
        self._account_id: int = self._infer_account_id()

    def _headers(self, content_type: ContentType) -> dict:
        return {
            "Accept": content_type.value,
            "Authorization": "Token " + self._credentials.token,
        }

    def _call(
        self,
        url: str,
        params: Optional[dict] = None,
        content_type: ContentType = ContentType.JSON,
    ) -> dict:
        headers = self._headers(content_type)
        response = self._session.get(
            url=url,
            headers=headers,
            params=params,
        )
        try:
            result = response.json()
        except:
            context = f"{url}, status {response.status_code}"
            raise ValueError(f"Couldn't extract data from {context}")
        else:
            if content_type == ContentType.JSON:
                return result[_DATA_KEY]
            return result

    def _infer_account_id(self) -> int:
        result = self._call(url=_CLOUD_URL)
        return result[0]["id"]

    def _last_run_id(self) -> int:
        """
        Extract last successful run id
        https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs/operation/listRunsForAccount
        """

        url = f"{_CLOUD_URL}/{self._account_id}/runs/"

        params = {
            "job_definition_id": self._credentials.job_id,
            "order_by": "-finished_at",
            "status": _SUCCESSFUL_RUN_STATUS,
            "limit": 1,
        }
        return self._call(url, params)[0]["id"]

    def fetch_manifest(self) -> dict:
        """
        Fetch dbt manifest
        https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs/operation/getArtifactsByRunId
        """
        run_id = self._last_run_id()
        url = f"{_CLOUD_URL}/{self._account_id}/runs/{run_id}/artifacts/{_MANIFEST_PATH}"
        logger.info(f"Extracting manifest from run id {run_id} with url {url}")

        # setting text content as a workaround to this issue
        # https://stackoverflow.com/questions/68201659/dbt-cloud-api-to-extract-run-artifacts
        return self._call(url, content_type=ContentType.TEXT)



from datetime import datetime, timedelta
from unittest.mock import ANY, MagicMock, Mock, call, patch

import pytest
from requests import HTTPError

from .constants import GET, POST, Assertions, Keys, QueryParams, Urls
from .credentials import Credentials
from .rest import Client, msal

FAKE_TENANT_ID = "IamFake"
FAKE_CLIENT_ID = "MeTwo"
FAKE_SECRET = "MeThree"


def _client() -> Client:
    creds = Credentials(
        tenant_id=FAKE_TENANT_ID,
        client_id=FAKE_CLIENT_ID,
        secret=FAKE_SECRET,
    )
    return Client(creds)


def _raise_http_error() -> None:
    raise HTTPError


@patch.object(msal, "ConfidentialClientApplication")
def test__access_token(mock_app):
    # init mocks
    valid_response = {"access_token": "mock_token"}
    returning_valid_token = Mock(return_value=valid_response)
    mock_app.return_value.acquire_token_silent = Mock(return_value=None)
    mock_app.return_value.acquire_token_for_client = returning_valid_token

    # init client
    client = _client()

    # generated token
    assert client._access_token() == valid_response

    # via silent endpoint token
    mock_app.return_value.acquire_token_silent = returning_valid_token
    mock_app.return_value.acquire_token_for_client = None
    assert client._access_token() == valid_response

    # token missing in response
    invalid_response = {"not_access_token": "666"}
    returning_invalid_token = Mock(return_value=invalid_response)
    mock_app.return_value.acquire_token_silent = returning_invalid_token

    with pytest.raises(ValueError):
        client._access_token()


@patch.object(Client, "_access_token")
def test__headers(mock_access_token):
    client = _client()
    mock_access_token.return_value = {Keys.ACCESS_TOKEN: "666"}
    assert client._header() == {"Authorization": "Bearer 666"}


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__get(mocked_access_token, mocked_request):
    client = _client()
    mocked_access_token.return_value = {Keys.ACCESS_TOKEN: "666"}
    fact = {"fact": "Approximately 24 cat skins can make a coat.", "length": 43}
    mocked_request.return_value = Mock(json=lambda: fact)

    result = client._get("https://catfact.ninja/fact")
    assert result == fact

    result = client._get("https://catfact.ninja/fact")["length"]
    assert result == 43

    mocked_request.return_value = Mock(raise_for_status=_raise_http_error)

    with pytest.raises(HTTPError):
        result = client._get("https/whatev.er")


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__workspace_ids(_, mocked_request):
    client = _client()
    mocked_request.return_value = Mock(
        json=lambda: [{"id": 1000}, {"id": 1001}, {"id": 1003}],
    )
    ids = client._workspace_ids()
    assert ids == [1000, 1001, 1003]

    with pytest.raises(AssertionError, match=Assertions.DATETIME_TOO_OLD):
        good_old_time = datetime(1998, 7, 12)
        client._workspace_ids(modified_since=good_old_time)

    yesterday = datetime.today() - timedelta(1)
    ids = client._workspace_ids(modified_since=yesterday)
    params = {
        Keys.INACTIVE_WORKSPACES: True,
        Keys.PERSONAL_WORKSPACES: True,
        Keys.MODIFIED_SINCE: f"{yesterday.isoformat()}0Z",
    }

    mocked_request.assert_called_with(
        GET,
        Urls.WORKSPACE_IDS,
        data=None,
        headers=ANY,
        params=params,
    )


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__post_default(_, mocked_request):
    client = _client()
    url = "https://estcequecestbientotleweekend.fr/"
    params = QueryParams.METADATA_SCAN
    data = {"bonjour": "hello"}
    client._post(url, params=params, data=data)
    mocked_request.assert_called_with(
        POST,
        url,
        headers=ANY,
        params=QueryParams.METADATA_SCAN,
        data=data,
    )


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__post_with_processor(_, mocked_request):
    client = _client()
    url = "https://estcequecestbientotleweekend.fr/"
    params = QueryParams.METADATA_SCAN
    data = {"bonjour": "hello"}
    mocked_request.return_value = Mock(json=lambda: {"id": 1000})
    result = client._post(
        url,
        params=params,
        data=data,
        processor=lambda x: x.json()["id"],
    )
    assert result == 1000


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__datasets(_, mocked_request):
    client = _client()
    mocked_request.return_value = Mock(
        json=lambda: {"value": [{"id": 1, "type": "dataset"}]},
    )
    datasets = client._datasets()
    mocked_request.assert_called_with(
        GET,
        Urls.DATASETS,
        data=None,
        headers=ANY,
        params=None,
    )
    assert datasets == [{"id": 1, "type": "dataset"}]


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__reports(_, mocked_request):
    client = _client()
    mocked_request.return_value = Mock(
        json=lambda: {"value": [{"id": 1, "type": "report"}]},
    )
    reports = client._reports()
    mocked_request.assert_called_with(
        GET,
        Urls.REPORTS,
        data=None,
        headers=ANY,
        params=None,
    )
    assert reports == [{"id": 1, "type": "report"}]


@patch("requests.request")
@patch.object(Client, "_access_token")
def test__dashboards(_, mocked_request):
    client = _client()
    mocked_request.return_value = Mock(
        json=lambda: {"value": [{"id": 1, "type": "dashboard"}]},
    )
    dashboards = client._dashboards()
    mocked_request.assert_called_with(
        GET,
        Urls.DASHBOARD,
        data=None,
        headers=ANY,
        params=None,
    )
    assert dashboards == [{"id": 1, "type": "dashboard"}]


@patch.object(Client, "_workspace_ids")
@patch.object(Client, "_create_scan")
@patch.object(Client, "_wait_for_scan_result")
@patch.object(Client, "_get_scan")
def test__metadata(
    mocked_get_scan,
    mocked_wait,
    mocked_create_scan,
    mocked_workspace_ids,
):
    mocked_workspace_ids.return_value = list(range(200))
    mocked_create_scan.return_value = 314
    mocked_wait.return_value = True
    mocked_get_scan.return_value = [{"workspace_id": 1871}]

    client = _client()
    result = client._metadata()

    assert list(result) == [[{"workspace_id": 1871}], [{"workspace_id": 1871}]]


_CALLS = [
    {
        Keys.ACTIVITY_EVENT_ENTITIES: ["foo", "bar"],
        Keys.LAST_RESULT_SET: False,
        Keys.CONTINUATION_URI: "https://next-call-1",
    },
    {
        Keys.ACTIVITY_EVENT_ENTITIES: ["baz"],
        Keys.LAST_RESULT_SET: False,
        Keys.CONTINUATION_URI: "https://next-call-2",
    },
    {
        Keys.ACTIVITY_EVENT_ENTITIES: ["biz"],
        Keys.LAST_RESULT_SET: True,
        Keys.CONTINUATION_URI: None,
    },
]


@patch.object(Client, "_call")
def test__activity_events(mocked):
    client = _client()
    mocked.side_effect = _CALLS

    result = client._activity_events()
    assert result == ["foo", "bar", "baz", "biz"]

    expected_calls = [
        call(ANY, GET, params=None, processor=None),
        call("https://next-call-1", GET, params=None, processor=None),
        call("https://next-call-2", GET, params=None, processor=None),
    ]
    mocked.assert_has_calls(expected_calls)



from datetime import date, datetime

from source.packages.extractor.castor_extractor.utils import at_midnight


def test__at_midnight():
    assert at_midnight(date(1989, 2, 2)) == datetime(1989, 2, 2, 0, 0, 0)



from typing import NamedTuple, Optional

from ...utils import OUTPUT_DIR, from_env, validate_baseurl
from .constant import ALL_LOOKS, BASE_URL, CLIENT_ID, CLIENT_SECRET


def _all_looks_from_env() -> bool:
    parameter = from_env(ALL_LOOKS, allow_missing=True)
    return str(parameter).lower() == "true"


class Parameters(NamedTuple):
    """Parameters for Looker extraction"""

    output_directory: str
    base_url: str
    client_id: str
    client_secret: str
    timeout: Optional[int]
    is_safe_mode: bool
    all_looks: bool


def get_parameters(**kwargs) -> Parameters:
    """
    Returns parameters for Looker extraction whether they come from script
    argument, env variable or default value.
    """
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)

    base_url = validate_baseurl(kwargs.get("base_url") or from_env(BASE_URL))
    client_id = kwargs.get("client_id") or from_env(CLIENT_ID)
    client_secret = kwargs.get("client_secret") or from_env(CLIENT_SECRET)

    # timeout can be set with env variable however we don't use from_env because it has a default value
    timeout = kwargs.get("timeout")

    is_safe_mode = kwargs.get("safe_mode", False)
    all_looks = kwargs.get("all_looks") or _all_looks_from_env()

    return Parameters(
        output_directory=output_directory,
        base_url=base_url,
        client_id=client_id,
        client_secret=client_secret,
        timeout=timeout,
        is_safe_mode=is_safe_mode,
        all_looks=all_looks,
    )



import logging
from typing import Iterable, Iterator, Optional, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import SigmaAsset
from .client import SigmaClient, SigmaCredentials
from .constants import API_TOKEN, CLIENT_ID, HOST

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: SigmaClient,
) -> Iterable[Tuple[SigmaAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Sigma"""

    logger.info("Extracting DATASETS from API")
    datasets = client.fetch(SigmaAsset.DATASETS)
    yield SigmaAsset.DATASETS, list(deep_serialize(datasets))

    logger.info("Extracting WORKBOOKS from API")
    workbooks = list(client.fetch(SigmaAsset.WORKBOOKS))
    yield SigmaAsset.WORKBOOKS, list(deep_serialize(workbooks))

    logger.info("Extracting FILES from API")
    files = client.fetch(SigmaAsset.FILES)
    yield SigmaAsset.FILES, list(deep_serialize(files))

    logger.info("Extracting MEMBERS from API")
    members = client.fetch(SigmaAsset.MEMBERS)
    yield SigmaAsset.MEMBERS, list(deep_serialize(members))

    logging.info("Extracting QUERIES data from API")
    queries = client.fetch(SigmaAsset.QUERIES, workbooks=workbooks)
    yield SigmaAsset.QUERIES, list(deep_serialize(queries))

    logging.info("Extracting ELEMENTS data from API")
    elements = list(client.fetch(SigmaAsset.ELEMENTS, workbooks=workbooks))
    yield SigmaAsset.ELEMENTS, list(deep_serialize(elements))

    logging.info("Extracting LINEAGES data from API")
    lineages = client.fetch(SigmaAsset.LINEAGES, elements=elements)
    yield SigmaAsset.LINEAGES, list(deep_serialize(lineages))


def extract_all(
    host: Optional[str] = None,
    client_id: Optional[str] = None,
    api_token: Optional[str] = None,
    output_directory: Optional[str] = None,
) -> None:
    """
    Extract data from Sigma API
    Store the output files locally under the given output_directory
    """

    _output_directory = output_directory or from_env(OUTPUT_DIR)
    _client_id = client_id or from_env(CLIENT_ID)
    _host = host or from_env(HOST)
    _api_token = api_token or from_env(API_TOKEN)

    credentials = SigmaCredentials(
        host=_host,
        client_id=_client_id,
        api_token=_api_token,
    )
    client = SigmaClient(credentials=credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts, host=_host)



from requests.exceptions import (
    ConnectionError as RequestConnectionError,
    Timeout as RequestTimeout,
)

from ....utils import RetryStrategy
from ..errors import RateLimitResponseError

RETRY_EXCEPTIONS = (
    RateLimitResponseError,
    ConnectionError,
    RequestConnectionError,
    RequestTimeout,
)
RETRY_COUNT = 2
RETRY_BASE_MS = 10_000
RETRY_JITTER_MS = 1_000
RETRY_STRATEGY = RetryStrategy.LINEAR


CLIENT_NAME = "ModeAnalytics/API"



import json
import logging
from collections import OrderedDict
from typing import cast

from ...logger import add_logging_file_handler
from ...utils import LocalStorage, SafeMode, from_env, write_summary
from ..abstract import (
    CATALOG_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    ExtractionProcessor,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
)
from .client import BigQueryClient
from .query import BigQueryQueryBuilder

logger = logging.getLogger(__name__)


BIGQUERY_CREDENTIALS = "GOOGLE_APPLICATION_CREDENTIALS"


BIGQUERY_ASSETS: SupportedAssets = OrderedDict(
    {
        WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
        WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
        WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
        WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    },
)


def _credentials(params: dict) -> dict:
    """extract GCP credentials"""
    path = params.get("credentials") or from_env(BIGQUERY_CREDENTIALS)
    logger.info(f"Credentials fetched from {path}")
    with open(path, "r") as file:
        return cast(dict, json.load(file))


def extract_all(**kwargs) -> None:
    """
    Extract all assets from BigQuery and store the results in CSV files
    Time filter scope for `Queries` = the day before (0 > 23h)
    """
    output_directory, skip_existing = common_args(kwargs)
    is_safe_mode = kwargs.get("safe_mode") or False

    safe_mode_params = None

    if is_safe_mode:
        add_logging_file_handler(output_directory)
        safe_mode_params = SafeMode((Exception,), float("inf"))

    client = BigQueryClient(
        credentials=_credentials(kwargs),
        db_allowed=kwargs.get("db_allowed"),
        db_blocked=kwargs.get("db_blocked"),
        dataset_blocked=kwargs.get("dataset_blocked"),
    )

    logger.info(f"Available projects: {client.get_projects()}\n")

    query_builder = BigQueryQueryBuilder(
        regions=client.get_regions(),
        datasets=client.get_datasets(),
    )

    storage = LocalStorage(directory=output_directory)

    extractor = ExtractionProcessor(
        client=client,
        query_builder=query_builder,
        storage=storage,
        safe_mode=safe_mode_params,
    )
    asset_count: int = 0
    for group in BIGQUERY_ASSETS.values():
        for asset in group:
            logger.info(f"Extracting `{asset.value.upper()}` ...")
            location = extractor.extract(asset, skip_existing)
            logger.info(f"Results stored to {location}\n")
            asset_count += 1

    logger.info(f"Extracted {asset_count} assets")
    if safe_mode_params:
        logger.info(f"{len(safe_mode_params.errors_caught)} assets skipped")

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from os import path as op


def load_file(path: str, calling_file: str) -> str:
    """Loads the given file using path relative to the calling file"""
    filepath = op.join(op.dirname(calling_file), path)
    with open(filepath, "r") as f:
        return f.read()



import logging
import os

from .utils import current_timestamp

MAIN_LOGGER_NAME = "castor_extractor"
CASTOR_LOG_FILENAME = "castor.log"
LOG_FORMAT = "%(levelname)s - %(message)s"


logger = logging.getLogger(MAIN_LOGGER_NAME)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(LOG_FORMAT)

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.INFO)
logger.addHandler(stream_handler)


def add_logging_file_handler(output_directory: str) -> None:
    """
    Add a filehandler to the main logger.
    """
    filename = f"{current_timestamp()}-{CASTOR_LOG_FILENAME}"
    file_path = os.path.join(output_directory, filename)

    fmt = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    file_handler = logging.FileHandler(file_path)
    file_handler.setFormatter(fmt)

    logger.addHandler(file_handler)



import unittest

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client.credentials import CredentialsApi


def test_default_site_id():
    creds = CredentialsApi(  # noqa: S106
        server_url="url",
        user="test_user",
        password="test_pwd",
        token="token",
        token_name="token_name",
        site_id=None,
    )
    assert creds.to_dict()["site_id"] == ""



import logging
import random
import time
from enum import Enum
from typing import Any, Callable, Optional, Sequence, Tuple, Type, Union
from urllib.error import HTTPError

logger = logging.getLogger(__name__)


MS_IN_SEC = 1000


class RetryStrategy(Enum):
    """retry mechanism"""

    CONSTANT = "CONSTANT"
    LINEAR = "LINEAR"
    EXPONENTIAL = "EXPONENTIAL"


DEFAULT_STRATEGY = RetryStrategy.CONSTANT


class Retry:
    """handling retry check and wait"""

    def __init__(
        self,
        count: int,
        base_ms: int,
        jitter_ms: int,
        strategy: Optional[RetryStrategy] = None,
    ):
        self._max = count
        self._base = base_ms
        self._jitter = jitter_ms
        self._strategy = strategy or DEFAULT_STRATEGY
        self.count = 0

    def jitter(self) -> int:
        """compute random jitter in ms"""
        min_ = self._jitter // 2
        max_ = int(self._jitter * 1.5)
        return random.randrange(min_, max_)

    def base(self) -> int:
        """compute base wait time in ms"""
        base = self._base
        if self._strategy == RetryStrategy.CONSTANT:
            return base
        if self._strategy == RetryStrategy.LINEAR:
            return base * self.count
        # exponential
        scaled = float(base) / MS_IN_SEC
        return int(scaled**self.count * MS_IN_SEC)

    def check(self) -> bool:
        """check wheter retry should happen or not"""
        if self.count >= self._max:
            return False
        self.count += 1
        wait_ms = self.base() + self.jitter()
        time.sleep(float(wait_ms) / MS_IN_SEC)
        return True


WrapperReturnType = Union[Tuple[BaseException, None], Tuple[None, Any]]


def retry(
    exceptions: Sequence[Type[BaseException]],
    count: int = 1,
    base_ms: int = 0,
    jitter_ms: int = 0,
    strategy: Optional[RetryStrategy] = None,
) -> Callable:
    """retry decorator"""

    exceptions_ = tuple(e for e in exceptions)

    def _wrapper(callable: Callable) -> Callable:
        def _try(*args, **kwargs) -> WrapperReturnType:
            try:
                return None, callable(*args, **kwargs)
            except exceptions_ as err:
                return err, None

        def _func(*args, **kwargs) -> Any:
            retry = Retry(count, base_ms, jitter_ms, strategy)
            while True:
                err, result = _try(*args, **kwargs)
                if err is None:
                    return result
                if retry.check():
                    logger.warning(f"retrying following: {err}")
                    continue
                raise err

        return _func

    return _wrapper



import os
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client import ApiClient
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "../assets/rest_api/")

GET_XML = os.path.join(TEST_ASSET_DIR, "view_get_usage.xml")

EXPECTED = [
    {
        "workbook_id": "3cc6cd06-89ce-4fdc-b935-5294135d6d42",
        "view_counts": 11,
    },
    {
        "workbook_id": "6d13b0ca-043d-4d42-8c9d-3f3313ea3a00",
        "view_counts": 13,
    },
]


class UsageTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient()
        self._client._server = TSC.Server("http://test_usage")

        # Fake signin
        self._client._server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self._client._server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self._client._server.views.baseurl

    def test_fetch_usages(self):
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(
                self.baseurl + "?includeUsageStatistics=true",
                text=response_xml,
            )
            results = self._client._fetch_usages(False)

            self.assertEqual(results, EXPECTED)



import logging
from typing import Optional

logger = logging.getLogger(__name__)


class Pagination:
    """This is a wrapper around Sigma's pagination system"""

    def __init__(
        self,
        next_page: Optional[str],
        entries: Optional[list] = None,
        total: Optional[int] = 0,
    ):
        self.next_page = next_page
        self.entries = entries or []
        self.total = total

    def generate_url(self, endpoint_url: str) -> str:
        """Generates the paginated url based on the targeted endpoint"""
        pagination = f"?page={self.next_page}"
        paginated_url = f"{endpoint_url}{pagination}"
        return paginated_url



from .uri import uri_encode


def test_uri_encode():
    assert uri_encode("Yok0On0Jon4L3nn0n_") == "Yok0On0Jon4L3nn0n_"
    assert uri_encode("Yok0On0:Jon4L3nn0n#") == "Yok0On0%3AJon4L3nn0n%23"
    assert uri_encode("password@robase") == "password%40robase"



from typing import Dict, Type

from .constants import ACCESS_DENIED_ERROR_CODE, APP_SIZE_EXCEEDED_ERROR_CODE


class JsonRpcError(Exception):
    """
    Error class to be raised when a JSON-RPC call response is an error
    """

    def __init__(self, message: dict, error: dict):
        self.message = message
        self.error = error

        name = self.__class__.__name__
        super().__init__({"type": name, "message": message, "error": error})


class AccessDeniedError(JsonRpcError):
    """
    Error class to be raised when JSON-RPC error is access denied
    """

    ...


class AppSizeExceededError(JsonRpcError):
    """
    Error class to be raised when JSON-RPC error is source size exceeded
    """

    ...


ERROR_CODE_MAPPING: Dict[int, Type[JsonRpcError]] = {
    ACCESS_DENIED_ERROR_CODE: AccessDeniedError,
    APP_SIZE_EXCEEDED_ERROR_CODE: AppSizeExceededError,
}


def raise_for_error(message: dict, response: dict) -> None:
    """Raises JsonRpcError when response is an error"""
    error = response.get("error")
    if not error:
        return None

    error_code = error["code"]
    exception = ERROR_CODE_MAPPING.get(error_code, JsonRpcError)
    raise exception(message, response["error"])



import inspect
import os
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple

from .asset import WarehouseAsset
from .time_filter import TimeFilter

TIME_FILTERED_ASSETS = (WarehouseAsset.COLUMN_LINEAGE, WarehouseAsset.QUERY)

QUERIES_DIR = "queries"


class AssetNotSupportedError(NotImplementedError):
    def __init__(self, asset: WarehouseAsset, builder_name: str):
        msg = f"Extraction of {asset} is not supported by {builder_name}"
        super().__init__(msg)


class ExtractionQuery:
    """contains both statement and parameters for the query"""

    def __init__(self, statement: str, params: dict):
        self.statement = statement
        self.params = params


class AbstractQueryBuilder(ABC):
    """
    Build queries necessary to extract warehouse assets.
    Handle feeding of parameters as well.
    """

    def __init__(
        self,
        time_filter: Optional[TimeFilter],
        duplicated: Optional[Tuple[WarehouseAsset, ...]] = None,
    ):
        self._time_filter = time_filter or TimeFilter.default()
        self._duplicated = duplicated

    def needs_deduplication(self, asset: WarehouseAsset) -> bool:
        if self._duplicated:
            return asset in self._duplicated
        return False

    def file_name(self, asset: WarehouseAsset) -> str:
        """Returns the SQL filename extracting the given asset"""
        return f"{asset.value}.sql"

    def _load_from_file(self, filename: str) -> str:
        """read from a file located in queries directory"""
        root = os.path.dirname(inspect.getfile(self.__class__))
        path = os.path.join(root, QUERIES_DIR, filename)
        with open(path, "r") as f:
            return f.read()

    def load_statement(self, asset: WarehouseAsset) -> str:
        """load SQL statement from file"""
        filename = self.file_name(asset)
        try:
            return self._load_from_file(filename)
        except FileNotFoundError:
            raise AssetNotSupportedError(asset, self.__class__.__name__)

    def build_default(self, asset: WarehouseAsset) -> ExtractionQuery:
        statement = self.load_statement(asset)
        params: dict = {}
        if asset in TIME_FILTERED_ASSETS:
            params = self._time_filter.to_dict()

        return ExtractionQuery(statement, params)

    @abstractmethod
    def build(self, asset: WarehouseAsset) -> List[ExtractionQuery]:
        """
        Build the Query allowing extraction of the given asset
        - Most of the time, returns a single query
        - Sometimes we must stitch several queries, hence the list
        """
        pass



from datetime import datetime, timedelta
from typing import List

from .constants import RECENT_DAYS, Assertions, Batches


def batch_size_is_valid_or_assert(ids: List) -> None:
    """
    assert that current batch is smaller than expected size
    """
    assert len(ids) <= Batches.METADATA, Assertions.BATCH_TOO_BIG


def datetime_is_recent_or_assert(dt: datetime) -> None:
    """
    assert that given datetime is recent
    """
    valid = dt > datetime.utcnow() - timedelta(RECENT_DAYS)
    assert valid, Assertions.DATETIME_TOO_OLD



import os
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client import ApiClient
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "../assets/rest_api/")

GET_XML = os.path.join(TEST_ASSET_DIR, "user_get.xml")

EXPECTED = [
    {
        "id": "dd2239f6-ddf1-4107-981a-4cf94e415794",
        "name": "alice",
        "email": "alicecook@test.com",
        "fullname": "alice cook",
        "site_role": "Publisher",
    },
    {
        "id": "2a47bbf8-8900-4ebb-b0a4-2723bd7c46c3",
        "name": "Bob",
        "email": "bob@test.com",
        "fullname": "Bob Smith",
        "site_role": "Interactor",
    },
]


class UserTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient()
        self._client._server = TSC.Server("http://test")

        # Fake signin
        self._client._server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self._client._server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self._client._server.users.baseurl

    def test_fetch_users(self):
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl + "?fields=_all_", text=response_xml)
            results = self._client._fetch_users()

        self.assertEqual(results, EXPECTED)



import logging
from typing import Iterable, Tuple

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_errors_logs,
    write_json,
    write_summary,
)
from .assets import TableauAsset
from .client import ApiClient as Client

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: Client,
) -> Iterable[Tuple[TableauAsset, list]]:
    """Iterate over the extracted Data from Tableau"""

    logger.info("Extracting USER from Tableau API")
    yield TableauAsset.USER, deep_serialize(client.fetch(TableauAsset.USER))

    logger.info("Extracting WORKBOOK from Tableau API")
    yield TableauAsset.WORKBOOK, deep_serialize(
        client.fetch(TableauAsset.WORKBOOK),
    )

    logger.info("Extracting PUBLISHED DATASOURCE from Tableau API")
    yield TableauAsset.PUBLISHED_DATASOURCE, deep_serialize(
        client.fetch(TableauAsset.PUBLISHED_DATASOURCE),
    )

    logger.info("Extracting PROJECT from Tableau API")
    yield TableauAsset.PROJECT, deep_serialize(
        client.fetch(TableauAsset.PROJECT),
    )

    logger.info("Extracting USAGE from Tableau API")
    yield TableauAsset.USAGE, deep_serialize(client.fetch(TableauAsset.USAGE))

    logger.info("Extracting WORKBOOK_TO_DATASOURCE from Tableau API")
    yield TableauAsset.WORKBOOK_TO_DATASOURCE, deep_serialize(
        client.fetch(TableauAsset.WORKBOOK_TO_DATASOURCE),
    )

    logger.info("Extracting DATASOURCE from Tableau API")
    yield TableauAsset.DATASOURCE, deep_serialize(
        client.fetch(TableauAsset.DATASOURCE),
    )

    logger.info("Extracting CUSTOM_SQL_TABLE from Tableau API")
    yield TableauAsset.CUSTOM_SQL_TABLE, deep_serialize(
        client.fetch(TableauAsset.CUSTOM_SQL_TABLE),
    )

    logger.info("Extracting CUSTOM_SQL_QUERY from Tableau API")
    yield TableauAsset.CUSTOM_SQL_QUERY, deep_serialize(
        client.fetch(TableauAsset.CUSTOM_SQL_QUERY),
    )

    logger.info("Extracting FIELD from Tableau API")
    yield TableauAsset.FIELD, deep_serialize(client.fetch(TableauAsset.FIELD))


def extract_all(client: Client, **kwargs: str) -> None:
    """
    Extract Data from tableau
    Store data locally in files under the output_directory
    If errors from Tableau's API are catch store them locally in file under the output_directory
    """
    output_directory = kwargs.get("output_directory") or from_env(OUTPUT_DIR)

    timestamp = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.value, output_directory, timestamp)
        write_json(filename, data)

    write_summary(
        output_directory,
        timestamp,
        base_url=client.base_url(),
        client_name=client.name(),
    )

    if client.errors:
        write_errors_logs(output_directory, timestamp, client.errors)
